local _,stash = ...
local L = stash.localization
local C = stash.constants
local settings = stash.settings
stash.info = {}

-- stash.info.items is a big table indexed by itemID of all stashed items {count,name,icon,rarity,type,subtype,equipslot}
-- on login this table is empty; but once items are gathered into it they will remain cached here
stash.info.items = {} 
-- stash.info.found is an ordered list of {realm,faction,player,location,count} for the found location/counts of a single itemID
stash.info.found = {} 

local searchMask = "" -- set by stash.info:SetSearch(), the desensitized text to search for
local isSearching = false -- true when searchMask is something other than ""

-- desensitize(text) returns text in a literal/case-insensitive format "[lL][iI][kK][eE] [tT][hH][iI][sS]%." for searches
local function literal(c) return "%"..c end
local function caseinsensitive(c) return format("[%s%s]",c:lower(),c:upper()) end
local function desensitize(text)
    return type(text)=="string" and text:trim():gsub("[%(%)%.%%%+%-%*%?%[%]%^%$]",literal):gsub("%a",caseinsensitive)
end

-- sets the searchMask to the given text (with special characters escaped and letters case insensitive)
-- returns true if the mask changed and new isSearching result as second return
function stash.info:SetSearch(text)
    local newMask = desensitize(text) or ""
    isSearching = newMask~=""
    if newMask==searchMask then
        return false, isSearching -- mask didn't change
    else
        searchMask = newMask -- mask changed
        return true, isSearching
    end
end

-- adds an individual itemID (and its count) to stash.info.items
function stash.info:AddItemID(itemID,count)
    if not stash.info.items[itemID] then
        stash.info.items[itemID] = {[C.INFO_COUNT]=0}
    end
    -- if doing a search and name isn't found, set count to 0 so it's not in the results
    if isSearching then
        local itemName = stash.info.items[itemID][C.INFO_NAME]
        if settings.SearchTypesToo then
            local itemType = stash.info.items[itemID][C.INFO_TYPE]
            local itemSubType = stash.info.items[itemID][C.INFO_SUBTYPE]
            local itemEquipSlot = stash.info.items[itemID][C.INFO_EQUIPSLOT]
            if not ((itemName and itemName:match(searchMask)) or (itemType and itemType:match(searchMask)) or (itemSubType and itemSubType:match(searchMask)) or (itemEquipSlot and itemEquipSlot:match(searchMask))) then
                count = 0
            end
        else -- only doing a search by name
            if itemName and not itemName:match(searchMask) then
                count = 0
            end
        end
    end
    -- if filtering, set count to 0 if filter not met
    if settings.FilterType and stash.info.items[itemID][C.INFO_TYPE]~=settings.FilterType then
        count = 0
    end
    if settings.FilterSubType and stash.info.items[itemID][C.INFO_SUBTYPE]~=settings.FilterSubType then
        count = 0
    end
    if settings.FilterEquipSlot and stash.info.items[itemID][C.INFO_EQUIPSLOT]~=settings.FilterEquipSlot then
        count = 0
    end
    -- finally add the count (if any!) to the total
    stash.info.items[itemID][C.INFO_COUNT] = stash.info.items[itemID][C.INFO_COUNT] + count
end

-- recursive function to add all itemIDs in the node and child nodes to stash.info.items
function stash.info:GatherItems(node)
    if type(node)=="table" then
        local continue = true
        if node.player then
            -- this is a character node (realm, faction, player filter available here)
            if settings.CurrentRealmOnly and (node.realm~=stash.current.realm or node.faction~=stash.current.faction) then
                continue = false -- CurrentRealmOnly enabled, only continue if this is current realm and faction
            end
            if settings.FilterRealm and node.realm~=settings.FilterRealm then
                continue = false -- realm filter enabled and this isn't the filtered realm
            end
            if settings.FilterCharacter and node.player~=settings.FilterCharacter then
                continue = false -- character filter enabled and this isn't the filtered character
            end
        end
        if continue then
            for k,v in pairs(node) do
                local skip = false
                -- if k=="worn" or "bags" or etc filter available here
                if type(v)=="table" and settings.FilterLocation and C.LOCATION_LOOKUP[k] and k~=C.LOCATION_REVERSE_LOOKUP[settings.FilterLocation] then
                    skip = true
                end
                if not skip then
                    local itemID,count
                    if type(k)=="string" and type(v)=="number" and v>0 and k:match("currency") then -- this is currency, unordered table
                        stash.info:AddItemID(k,v)
                    elseif type(v)~="table" and type(k)=="number" then -- this is an itemID
                        if type(v)=="string" then -- can be an item:count or battlepet:speciesID
                            itemID,count = v:match("^(%d+):(%d+)$")
                            stash.info:AddItemID(tonumber(itemID) or v,tonumber(count) or 1)
                        else -- this is (likely) an itemID without a count
                            stash.info:AddItemID(v,1)
                        end
                    elseif type(v)=="table" then
                        stash.info:GatherItems(v) -- found a subtable, gather it too
                    end
                end
            end
        end
    end
end

-- goes through stash.info.items and fetches the name, rarity, type, subtype and equipslot of all itemIDs
function stash.info:CacheItemInfo()
    local isAnyNotCached = false
    for itemID,itemInfo in pairs(stash.info.items) do
        if not itemInfo[C.INFO_NAME] then -- this item has no name yet
            if type(itemID)=="string" then -- this is likely a battlepet
                local stringType,stringID = itemID:match("(%a+):(%d+)")
                --local speciesID = itemID:match("battlepet:(%d+)")
                if stringType=="battlepet" then
                    local name,icon,petType = C_PetJournal.GetPetInfoBySpeciesID(tonumber(stringID))
                    itemInfo[C.INFO_NAME] = name
                    itemInfo[C.INFO_ICON] = icon
                    itemInfo[C.INFO_RARITY] = 1 -- making all pets common quality (sorry pets!)
                    itemInfo[C.INFO_TYPE] = "Caged Battle Pet"
                    itemInfo[C.INFO_SUBTYPE] = _G["BATTLE_PET_NAME_"..petType]
                elseif stringType=="currency" then
                    local info = C_CurrencyInfo.GetBasicCurrencyInfo(tonumber(stringID))
                    itemInfo[C.INFO_NAME] = info.name
                    itemInfo[C.INFO_ICON] = info.icon
                    itemInfo[C.INFO_RARITY] = info.quality
                    itemInfo[C.INFO_TYPE] = L["Currency"]
                end
            elseif type(itemID)=="number" then -- this is an actual itemID
                if C_Item.IsItemDataCachedByID(itemID) then
                    local name,_,rarity,_,_,itemType,subType,_,equipslot,icon = GetItemInfo(itemID)
                    itemInfo[C.INFO_NAME] = name
                    itemInfo[C.INFO_ICON] = icon
                    itemInfo[C.INFO_RARITY] = rarity
                    itemInfo[C.INFO_TYPE] = itemType
                    itemInfo[C.INFO_SUBTYPE] = subType
                    itemInfo[C.INFO_EQUIPSLOT] = _G[equipslot] or equipslot
                else
                    C_Item.RequestLoadItemDataByID(itemID)
                    isAnyNotCached = true
                end
            end 
        end
    end
    -- if something wasn't cached, a request was made. come back in a bit to re-gather missing data
    if isAnyNotCached then
        C_Timer.After(C.DATA_CACHE_DELAY,stash.info.CacheItemInfo)
    end
end

function stash.info:GatherAllItems()
    for itemID,itemInfo in pairs(stash.info.items) do
        itemInfo[C.INFO_COUNT] = 0 -- reset count on all items
    end
    stash.info:GatherItems(stash.saved) -- kick off recursive gathering of itemIDs and their counts
    stash.info:CacheItemInfo() -- after itemIDs and counts gathered, gather names, types, etc for all items
end

-- returns the total money (copper) owned by all characters; total is true if it should ignore CurrentRealmOnly
function stash.info:CountAllMoney(total)
    local money = 0
    for realm,realmInfo in pairs(stash.saved) do
        for _,playerInfo in pairs(realmInfo) do
            if total or not settings.CurrentRealmOnly or (stash.current.realm==realm and stash.current.faction==playerInfo.faction) then
                money = money + playerInfo.money
            end
        end
    end
    return money
end

-- returns true if everything is data cached
function stash.info:IsEverythingDataCached()
    for _,info in pairs(self.items) do
        if not info[C.INFO_NAME] then
            return false -- something is missing a name, not everything data cached
        end
    end
    return true
end

-- adds a found item to stash.info.found
function stash.info:AddFound(realm,faction,player,location,count)
    if location~="money" and settings.CurrentRealmOnly and (realm~=stash.current.realm or faction~=stash.current.faction) then
        return -- if only looking at current realm and faction, ignore stuff found elsewhere
    end
    -- uncomment the following to filter tooltip found locations to match the active location filters
    -- if settings.FilterRealm and realm~=settings.FilterRealm then
    --     return -- realm filter enabled and this isn't the filtered realm
    -- end
    -- if settings.FilterCharacter and player~=settings.FilterCharacter then
    --     return -- character filter enabled and this isn't the filtered character
    -- end
    -- if settings.FilterLocation and location~=C.LOCATION_REVERSE_LOOKUP[settings.FilterLocation] then
    --     return -- location filter enabled and this isn't the filtered location
    -- end

    -- look for an existing found entry to add its count to
    for _,found in ipairs(stash.info.found) do
        if found[C.FOUND_REALM]==realm and found[C.FOUND_FACTION]==faction and found[C.FOUND_PLAYER]==player and found[C.FOUND_LOCATION]==location then
            found[C.FOUND_COUNT] = found[C.FOUND_COUNT] + count
            return
        end
    end
    -- if reached here, there wasn't an existing stash.info.found for this realm+faction+player+location
    tinsert(stash.info.found,{realm,faction,player,location,count})
end

-- used by FindItem, this is a recursive search to find the realm,faction,player,location that an item is found.
-- the results are stored in stash.info.found
local foundRealm, foundFaction, foundPlayer, foundLocation -- used to remember where an item is in the recursive search
function stash.info:FindItemByNode(node,itemID)
    if type(node)=="table" then
        if node.player then
            foundRealm = node.realm
            foundFaction = node.faction
            foundPlayer = node.player
        end
        for location,content in pairs(node) do
            if C.LOCATIONS[location] then -- if this is worn, bags, bank, reagents, then save it
                foundLocation = location
            end
            if type(location)=="string" and type(content)=="number" and location==itemID and location:match("^currency") then -- this is currency, unordered table
                stash.info:AddFound(foundRealm,foundFaction,foundPlayer,foundLocation,content)
            elseif type(content)~="table" and type(location)=="number" then
                -- we're inside a subtable slot like [worn][1] or [bags][-3][50]
                if content==itemID then
                    stash.info:AddFound(foundRealm,foundFaction,foundPlayer,foundLocation,1)
                    -- found it!
                elseif type(itemID)=="number" and type(content)=="string" then
                    local foundItemID,count = content:match("^(%d+):(%d+)$")
                    if tonumber(foundItemID)==itemID then
                        stash.info:AddFound(foundRealm,foundFaction,foundPlayer,foundLocation,tonumber(count))
                    end
                end
            elseif type(content)=="table" then
                stash.info:FindItemByNode(content,itemID)
            end
        end
    end
end

-- stash.info.SortFound is intended to be used in a table.sort to sort stash.info.found, either the locations of a single itemID
-- if sortFoundMode is C.SORT_FOUND_LOCATIONS, or for money if sortFoundMode is C.SORT_FOUND_MONEY
-- normal sort order is realm (current first then alpha) > faction (current first then alpha) > name > location
-- money sort order is realm (current first then alpha) > faction (current first then alpha) > money desc > name
local currentRealm=GetRealmName() -- these will be sorted to the top
local currentFaction=UnitFactionGroup("player")
local currentPlayer=UnitName("player")
local sortFoundMode = C.SORT_FOUND_LOCATIONS -- either C.SORT_FOUND_LOCATIONS or C.SORT_FOUND_MONEY
function stash.info.SortFound(found1,found2)
    if found1 and not found2 then
        return true
    elseif not found1 and found2 then
        return false
    elseif found1[C.FOUND_REALM]==currentRealm and found2[C.FOUND_REALM]~=currentRealm then -- current realm sorts first
        return true
    elseif found1[C.FOUND_REALM]~=currentRealm and found2[C.FOUND_REALM]==currentRealm then
        return false
    elseif found1[C.FOUND_REALM]~=found2[C.FOUND_REALM] then -- after current realm, sort by realm alphabetically
        return found1[C.FOUND_REALM] < found2[C.FOUND_REALM]
    elseif found1[C.FOUND_FACTION]==currentFaction and found2[C.FOUND_FACTION]~=currentFaction then -- current faction sorts first
        return true
    elseif found1[C.FOUND_FACTION]~=currentFaction and found2[C.FOUND_FACTION]==currentFaction then
        return false
    elseif found1[C.FOUND_FACTION]~=found2[C.FOUND_FACTION] then -- after current faction, sort by faction (if there's a neutral pandaren in the mix)
        return found1[C.FOUND_FACTION] < found2[C.FOUND_FACTION]
    elseif found1[C.FOUND_PLAYER]:match(" ") and not found2[C.FOUND_PLAYER]:match(" ") then -- if name has a space in it, sort to top after faction (this is a "Realm Faction" name)
        return true
    elseif not found1[C.FOUND_PLAYER]:match(" ") and found2[C.FOUND_PLAYER]:match(" ") then
        return false
    elseif sortFoundMode==C.SORT_FOUND_MONEY and found1[C.FOUND_COUNT]~=found2[C.FOUND_COUNT] then
        return found1[C.FOUND_COUNT] > found2[C.FOUND_COUNT]
    elseif found1[C.FOUND_PLAYER]==currentPlayer and found2[C.FOUND_PLAYER]~=currentPlayer then -- current player sorts first
        return true
    elseif found1[C.FOUND_PLAYER]~=currentPlayer and found2[C.FOUND_PLAYER]==currentPlayer then
        return false
    elseif found1[C.FOUND_PLAYER]~=found2[C.FOUND_PLAYER] then -- after current player, sort by player
        return found1[C.FOUND_PLAYER] < found2[C.FOUND_PLAYER]
    else
        return found1[C.FOUND_LOCATION] < found2[C.FOUND_LOCATION] -- finally sort by location
    end
end

-- populates and returns stash.info.found with the locations that an item is found
function stash.info:FindItem(itemID)
    wipe(stash.info.found)
    stash.info:FindItemByNode(stash.saved,itemID)
    sortFoundMode = C.SORT_FOUND_LOCATIONS
    table.sort(stash.info.found,stash.info.SortFound)
    return stash.info.found
end

-- populates and returns stash.info.found with the money for each character
function stash.info:FindMoney()
    wipe(stash.info.found)
    for realm,realmInfo in pairs(stash.saved) do
        for player,playerInfo in pairs(realmInfo) do
            tinsert(stash.info.found,{playerInfo.realm,playerInfo.faction,player,"money",playerInfo.money})
            stash.info:AddFound(realm,playerInfo.faction,format("%s %s",realm,playerInfo.faction),"money",playerInfo.money)
        end
    end
    sortFoundMode = C.SORT_FOUND_MONEY
    table.sort(stash.info.found,stash.info.SortFound)
    return stash.info.found
end

local encounteredTypes = {} -- unordered table indexed by type with unordered subtypes (which are tables to unordered equiplocs)
local encounteredLocations = {} -- unordered trable indexed by realm with unordered characters (which are tables to unordered locations)

-- updates encounteredTypes unordered table with encountered types/subtypes; note: no types/subtypes are ever removed
function stash.info:UpdateEncounteredTypes()
    for _,info in pairs(self.items) do
        local itemType = info[C.INFO_TYPE]
        if itemType and itemType~="" and not encounteredTypes[itemType] then
            encounteredTypes[itemType] = {}
        end
        local subType = info[C.INFO_SUBTYPE]
        if subType and subType~="" and not encounteredTypes[itemType][subType] then
            encounteredTypes[itemType][subType] = {}
        end
        local slot = info[C.INFO_EQUIPSLOT]
        if slot and slot~="" then
            encounteredTypes[itemType][subType][slot] = true
        end
    end
    return encounteredTypes
end

-- updates encounteredLocations unordered table with encountered realms/players/locations; note: no node is ever removed here
-- if a character is reset the encounteredLocations table should be wiped
function stash.info:UpdateEncounteredLocations()
    for realm,realmInfo in pairs(stash.saved) do
        if not encounteredLocations[realm] then
            encounteredLocations[realm] = {}
        end
        for player,playerInfo in pairs(realmInfo) do
            if not encounteredLocations[realm][player] then
                encounteredLocations[realm][player] = {}
            end
            for location,locationInfo in pairs(playerInfo) do
                if type(locationInfo)=="table" and not self:IsSubTablesEmpty(locationInfo) then -- only adding non-empy locations
                    encounteredLocations[realm][player][location] = true
                end
            end
        end
    end
    return encounteredLocations
end

-- if any character reset, then wipe encountered locations and rebuild
function stash.info:ResetEncounteredLocations()
    wipe(encounteredLocations)
end

-- returns an ordered list of types (if no args provided), subtypes of itemType if provided, and equipslots of subType if provided
function stash.info:GetTypes(itemType,subType)
    self:UpdateEncounteredTypes()
    local results = {} -- list compiled here; this is only used in dropdownmenus so a little garbage creation is ok (verify)
    local subList -- reference to the node in encounteredTypes we're getting the list from
    if not itemType then -- want a list of all C.INFO_TYPE across all items
        subList = encounteredTypes
    elseif not subType then -- want a list of all C.INFO_SUBTYPE for the given itemType
        subList = encounteredTypes[itemType]
    else  -- want a list of all C.INFO_EQUIPSLOT with the given itemType and subType
        subList = encounteredTypes[itemType] and encounteredTypes[itemType][subType]
    end
    -- subList is now an ordered list of the types/subtypes/equiplocs we're looking for
    if subList and type(subList)=="table" then
        for typeName in pairs(subList) do
            tinsert(results,typeName)
        end
        table.sort(results)
    end
    if #results>1 then -- only return a list if there's more than one
        return results
    else -- for everything else return an empty list
        return {}
    end
end

-- similar to GetTypes, but instead of returning a list, returns true/false if a list would exist (does this without creating
-- a table, so less garbage creation)
function stash.info:HasTypes(itemType,subType)
    self:UpdateEncounteredTypes()
    local count = 0
    local subList -- reference to the node in encounteredTypes we're getting the list from
    if not itemType then -- want a list of all C.INFO_TYPE across all items
        subList = encounteredTypes
    elseif not subType then -- want a list of all C.INFO_SUBTYPE for the given itemType
        subList = encounteredTypes[itemType]
    else  -- want a list of all C.INFO_EQUIPSLOT with the given itemType and subType
        subList = encounteredTypes[itemType] and encounteredTypes[itemType][subType]
    end
    if subList and type(subList)=="table" then
        for typeName in pairs(subList) do
            count = count + 1
            if count > 1 then
                return true -- there's more than 1 item type, enough to make a list
            end
        end
    end
    return false -- if made it here, there aren't enough items to make a list
end

-- returns an ordered list of realms (if no args provided), characters of realm if provided, and locations on character if provided
function stash.info:GetLocations(realm,character)
    self:UpdateEncounteredLocations()
    local results = {}
    local subList

    if not realm then
        subList = encounteredLocations
    elseif not character then
        subList = encounteredLocations[realm]
    else
        subList = encounteredLocations[realm] and encounteredLocations[realm][character]
    end
    if subList and type(subList)=="table" then
        for name in pairs(subList) do
            if realm and character and C.LOCATIONS[name] then -- if this is bags, bank, worn, etc; translate to a displayed name
                tinsert(results,C.LOCATIONS[name].name)
            else
                tinsert(results,name)
            end
        end
        table.sort(results)
    end
    -- locations will return lists with only one element too; since people can play on one realm or have one character on a realm
    return results
end

-- returns true if the given table and its immediate sub-tables are empty (used to discard unused locations like reagents on classic
-- or empty banks that a character hasn't opened yet with the addon enabled)
function stash.info:IsSubTablesEmpty(t)
    if type(t)~="table" then
        return
    end
    for k,v in pairs(t) do
        if type(v)~="table" then
            return false
        elseif next(v) then
            return false
        end
    end
    return true -- if made it here, there was nothing in the given table or its subtables
end

-- returns the faction of the given realm and character
function stash.info:GetFaction(realm,character)
    if realm and character and stash.saved[realm] and stash.saved[realm][character] then
        return stash.saved[realm][character].faction
    end
end

-- returns true if a search is being used
function stash.info:IsSearching()
    return searchMask:trim():len()>0
end